"""Clapeyron writers functions module."""

from .critical import write_critical
from .dortmund_groups import write_dortmund
from .molar_mass import write_molar_mass
from .psrk_groups import write_psrk
from .unifac_groups import write_unifac


__all__ = [
    "write_critical",
    "write_dortmund",
    "write_molar_mass",
    "write_psrk",
    "write_unifac",
]



"""Predictive Soave-Redlich-Kwong (PSRK) FragmentationModel implementation.

Import and use the PSRK FragmentationModel with:

.. code-block:: python

    from ugropy import psrk

    # Get groups from molecule's name
    tol = psrk.get_groups("toluene")

    print(tol.subgroups)

    # Get groups from molecule's SMILES
    eth = psrk.get_groups("CCO", "smiles")

    print(eth.subgroups)

Attributes
----------
psrk: GibbsModel
    Predictive Soave-Redlich-Kwong FragmentationModel :cite:p:`ddbst, psrk1,
    psrk2`
"""

from ugropy.constants import _csvs
from ugropy.core.frag_classes.gibbs_model.gibbs_model import GibbsModel
from ugropy.models.read_csv import _rd


# =============================================================================
# PSRK
# =============================================================================
_psrk = _csvs / "psrk"

_psrk_sg = _rd(_psrk / "psrk_subgroups.csv", "group")
_psrk_info = _rd(_psrk / "psrk_info.csv", "group")

psrk = GibbsModel(subgroups=_psrk_sg, subgroups_info=_psrk_info)



"""Core module.

The main algorithm functions and classes are included in this module.

- ilp_solvers:
    Module with the structure of the integer linear programming (ILP) solvers.
- frag_classes:
    Module with the classes that define the fragmentation models.
- check_atoms_fragments_presence:
    Check function to corroborate if ILP algorithm is necessary to solve a
    molecule.
- instantiate_mol_object:
    Function to instantiate a RDKit molecule object from a SMILES string, name
    or rdkit mol object.
"""

from . import frag_classes, ilp_solvers
from .checks import check_atoms_fragments_presence
from .get_rdkit_object import instantiate_mol_object


__all__ = [
    "ilp_solvers",
    "frag_classes",
    "check_atoms_fragments_presence",
    "instantiate_mol_object",
]



"""Read csv file with models' data auxiliar function."""

import pandas as pd


def _rd(file_path: str, index_col: str = None) -> pd.DataFrame:
    """Read the models' csv.

    Parameters
    ----------
    file_path : str or pathlib
        Path to csv file.
    index_col : str, optional
        Name of the index column, by default None.

    Returns
    -------
    pd.DataFrame
        Readed csv.
    """
    with open(file_path, mode="r") as f:
        return pd.read_csv(f, sep="|", index_col=index_col, comment="?")



"""ilp_solvers module.

This module contains the integer linear programming solvers used by `ugropy`.
ILP solvers are neccesary to solve the `Set Cover` problem, which is solved
to select the optimal set of fragments (groups) to cover the overlapping atoms
on a molecule.
"""

from . import default_solver, ilp_solver


__all__ = ["ilp_solver", "default_solver"]



"""to_yaeos module."""

from typing import List

from ugropy.core.frag_classes.gibbs_model.gibbs_model import GibbsModel
from ugropy.writers.thermo import to_thermo


def to_yaeos(mol_subgroups_list: List[dict], model: GibbsModel) -> str:
    """Obtain the Fortran source code for yaeos groups definition.

    yaeos: https://github.com/ipqa-research/yaeos

    Parameters
    ----------
    mol_subgroups_list : List[dict]
        List of ugropy subgroups dictionaries.
    model : GibbsModel
        Gibbs excess FragmentationModel (unifac, psrk, etc).

    Returns
    -------
    str
        Yaeos Fortran source code.
    """
    n_mol = len(mol_subgroups_list)

    code = (
        "use yaeos__models_ge_group_contribution_unifac, only: Groups\n"
        "\n"
        f"type(Groups) :: molecules({n_mol})\n"
        "\n"
    )

    # Obtain the dictionary of subgroups for each molecule but with index
    for i, mol_subgroups in enumerate(mol_subgroups_list):
        thermo_groups = to_thermo(mol_subgroups, model)

        molecule_code = (
            f"molecules({i+1})%groups_ids = "
            f"{list(thermo_groups.keys())}\n"
            f"molecules({i+1})%number_of_groups = "
            f"{list(thermo_groups.values())}\n"
            "\n"
        )

        code += molecule_code

    return code



"""check module.

The module contains the necessary checks to corroborate the success of the
algorithm to obtain the molecule's FragmentationModel subgroups.

"""

import numpy as np

from rdkit import Chem


def check_atoms_fragments_presence(
    molecule: Chem.rdchem.Mol, fragments: dict
) -> tuple[bool, np.ndarray]:
    """Find overlapped atoms and free atoms.

    Check the detected fragments to find the atoms that appears in more than
    one fragment (overlapping), and the atoms that are not present in any
    fragment (free atoms). Returning two np.ndarray with the indexes of the
    overlapping and free atoms.

    Example of a `fragments` dictionary that not presents overlapping atoms.
    For example N-hexane:

    .. code-block:: python

        {
            'CH3_0': (0,),
            'CH3_1': (5,),
            'CH2_0': (1,),
            'CH2_1': (2,),
            'CH2_2': (3,),
            'CH2_3': (4,)
        }

    Example of a `fragments` dictionary that presents overlapping atoms. For
    example toluene:

    .. code-block:: python

        {
            'CH3_0': (0,),
            'ACH_0': (2,),
            'ACH_1': (3,),
            'ACH_2': (4,),
            'ACH_3': (5,),
            'ACH_4': (6,),
            'AC_0': (1,),
            'ACCH3_0': (1, 0)
        }

    Parameters
    ----------
    molecule : Chem.rdchem.Mol
        RDKit molecule object.
    fragments : dict
        Dictionary containing the fragments detected in the molecule. The keys
        are the group names and the values are the indexes of the atoms in the
        group.

    Returns
    -------
    tuple[np.ndarray, np.ndarray]
        Overlapping atoms indexes and free atoms indexes.
    """
    n_atoms = molecule.GetNumAtoms()

    # Count the number of times an atom is in a group. Also find the atoms
    # that are not present in any fragment.
    atoms = np.zeros(n_atoms, dtype=int)

    for indexes in fragments.values():
        np.add.at(atoms, np.array(indexes).flatten(), 1)

    overlapped_atoms = np.argwhere(atoms > 1).flatten()
    free_atoms = np.argwhere(atoms == 0).flatten()

    return overlapped_atoms, free_atoms


# def check_has_molecular_weight_right(
#     self, mol_object: Chem.rdchem.Mol, fragments: dict
# ) -> bool:

#     # rdkit molecular weight
#     rdkit_mw = Descriptors.MolWt(mol_object)

#     # Molecular weight from functional groups
#     mws = self.mol_subgroups.loc[
#         list(fragments.keys()), "molecular_weight"
#     ].to_numpy()

#     func_group_mw = np.dot(mws, list(mol_subgroups.values()))

#     return np.allclose(rdkit_mw, func_group_mw, atol=0.5)



"""FragmentationModel module.

All ugropy models (joback, unifac, psrk, etc) are instances of the
FragmentationModule class.
"""

from collections import defaultdict
from typing import List, Union

import numpy as np

import pandas as pd

from rdkit import Chem
from rdkit.Chem import rdMolEnumerator

from ugropy.core.checks import check_atoms_fragments_presence
from ugropy.core.frag_classes.base.fragmentation_result import (
    FragmentationResult,
)
from ugropy.core.get_rdkit_object import instantiate_mol_object
from ugropy.core.ilp_solvers.default_solver import DefaultSolver
from ugropy.core.ilp_solvers.ilp_solver import ILPSolver


class FragmentationModel:
    """FragmentationModel class.

    All ugropy supported models are an instance of this class. This class must
    be inherited to create a new type of FragmentationModel.

    Parameters
    ----------
    subgroups : pd.DataFrame
        Model's subgroups. Index: 'group' (subgroups names). Mandatory columns:
        'smarts' (SMARTS representations of the group to detect its precense in
        the molecule).
    allow_overlapping : bool, optional
        Weather allow overlapping or not, by default False
    fragmentation_result : FragmentationResult, optional
        Fragmentation result class, by default FragmentationResult

    Attributes
    ----------
    subgroups : pd.DataFrame
        Model's subgroups. Index: 'group' (subgroups names). Mandatory columns:
        'smarts' (SMARTS representations of the group to detect its precense in
        the molecule).
    detection_mols : dict
        Dictionary cotaining all the rdkit Mol object from the detection_smarts
        subgroups column.
    """

    def __init__(
        self,
        subgroups: pd.DataFrame,
        allow_overlapping: bool = False,
        allow_free_atoms: bool = False,
        fragmentation_result: FragmentationResult = FragmentationResult,
    ) -> None:
        self.subgroups = subgroups
        self.allow_overlapping = allow_overlapping
        self.allow_free_atoms = allow_free_atoms
        self.fragmentation_result = fragmentation_result

        # Instantiate all de mol object from their SMARTS representation
        self.detection_mols = {}

        for group, row in self.subgroups.iterrows():
            smarts = row["smarts"]

            if "LN:" in smarts:
                # Linknode expression from Chemaxon SMARTS
                smarts = smarts.replace("LN:", "|LN:") + "|"
                query = Chem.MolFromSmarts(smarts)
                enumerated_smarts = rdMolEnumerator.Enumerate(query)
                self.detection_mols[group] = enumerated_smarts
            else:
                self.detection_mols[group] = Chem.MolFromSmarts(smarts)

    def get_groups(
        self,
        identifier: Union[str, Chem.rdchem.Mol],
        identifier_type: str = "name",
        solver: ILPSolver = DefaultSolver,
        search_multiple_solutions: bool = False,
        search_nonoptimal: bool = False,
        solver_arguments: dict = {},
        **kwargs,
    ) -> Union[FragmentationResult, List[FragmentationResult]]:
        """Get the groups of a molecule.

        Parameters
        ----------
        identifier : Union[str, Chem.rdchem.Mol]
            Identifier of the molecule. You can use either the name of the
            molecule, the SMILEs of the molecule or a rdkit Mol object.
        identifier_type : str, optional
            Identifier type of the molecule. Use "name" if you are providing
            the molecules' name, "smiles" if you are providing the SMILES or
            "mol" if you are providing a rdkir mol object, by default "name"
        solver : ILPSolver, optional
            ILP solver class, by default DefaultSolver
        search_multiple_solutions : bool, optional
            Weather search for multiple solutions or not, by default False If
            False the return will be a FragmentationResult object, if True the
            return will be a list of FragmentationResult objects.
        search_nonoptimal : bool, optional
            If True, the solver will search for non-optimal solutions along
            with the optimal ones. This is useful when the user wants to find
            all possible combinations of fragments that cover the universe. By
            default False. If `search_multiple_solutions` is False, this
            parameter will be ignored.
        solver_arguments : dict, optional
            Dictionary with the arguments to be passed to the solver. For the
            DefaultSolver of ugropy you can change de PulP solver passing a
            dictionary like {"solver": "PULP_CBC_CMD"} and change the PulP
            solver. If empty it will use the default solver arguments, by
            default {}.

        Returns
        -------
        Union[FragmentationResult, List[FragmentationResult]]
            Fragmentation result. If search_multiple_solutions is False the
            return will be a FragmentationResult object, if True the return
            will be a list of FragmentationResult objects.
        """
        # =====================================================================
        # Direct fragments detection
        # =====================================================================
        mol = self.mol_preprocess(
            instantiate_mol_object(identifier, identifier_type)
        )

        detections = self.detect_fragments(mol)

        # No groups detected
        if detections == {}:
            return self.set_fragmentation_result(
                mol, [{}], search_multiple_solutions, **kwargs
            )

        # =====================================================================
        # Search for overlapping atoms and free atoms
        # =====================================================================
        overlapping_atoms, free_atoms = check_atoms_fragments_presence(
            mol, detections
        )

        # If there is free atoms in the molecule can't fragment with the model
        if np.size(free_atoms) > 0 and not self.allow_free_atoms:
            return self.set_fragmentation_result(
                mol, [{}], search_multiple_solutions, **kwargs
            )

        # If no overlapping or the model allows overlapping we are done
        if np.size(overlapping_atoms) == 0 or self.allow_overlapping:
            return self.set_fragmentation_result(
                mol, [detections], search_multiple_solutions, **kwargs
            )

        # =====================================================================
        # Solve overlapping atoms
        # =====================================================================
        problem: ILPSolver = solver(
            overlapping_atoms,
            detections,
            search_multiple_solutions,
            search_nonoptimal,
            solver_arguments,
        )

        problem.solve()

        if not problem.selected_fragments:
            # This could happend, no solution found. Example:
            # "CC(C)(C)OC(=O)OC1=CC=CC=C1" on UNIFAC.
            return self.set_fragmentation_result(
                mol, [{}], search_multiple_solutions, **kwargs
            )

        solutions = []

        for selection in problem.selected_fragments:
            solution = detections.copy()

            for frag in problem.overlapped_fragments:
                if frag not in selection:
                    solution.pop(frag)

            solutions.append(solution)

        return self.set_fragmentation_result(
            mol, solutions, search_multiple_solutions, **kwargs
        )

    def mol_preprocess(self, mol: Chem.rdchem.Mol) -> Chem.rdchem.Mol:
        """Preprocess the molecule.

        This method is called before the detection of the fragments. It can be
        used to preprocess the molecule before the detection of the fragments.
        This allow to use RDKit functions to preprocess the molecule and make
        your SMARTS detection easier. The default implementation does nothing
        and leave the mol object as it is. You can inherit this class and
        override this method to preprocess the molecule.

        Parameters
        ----------
        mol : Chem.rdchem.Mol
            Molecule to preprocess.

        Returns
        -------
        Chem.rdchem.Mol
            Preprocessed molecule.
        """
        return mol

    def set_fragmentation_result(
        self,
        molecule: Chem.rdchem.Mol,
        solutions_fragments: List[dict],
        search_multiple_solutions: bool = False,
        **kwargs,
    ) -> Union[FragmentationResult, List[FragmentationResult]]:
        """Process the solutions and return the FragmentationResult objects.

        Parameters
        ----------
        molecule : Chem.rdchem.Mol
            Rdkit mol object.
        solutions_fragments : List[dict]
            Fragments detected in the molecule.
        search_multiple_solutions : bool, optional
            Weather search for multiple solutions or not, by default False

        Returns
        -------
        Union[FragmentationResult, List[FragmentationResult]]
            List of FragmentationResult objects.
        """
        sols = []
        occurs = []  # solutions registry to avoid repeated solutions

        for solution in solutions_fragments:
            occurrences = defaultdict(int)
            groups_atoms = defaultdict(list)

            for frag, atoms in solution.items():
                name = frag.split("_")[0]
                occurrences[name] += 1
                groups_atoms[name].append(atoms)

            if occurrences not in occurs:
                sols.append(
                    self.fragmentation_result(
                        molecule,
                        dict(occurrences),
                        dict(groups_atoms),
                        **kwargs,
                    )
                )
                occurs.append(occurrences)

        if search_multiple_solutions:
            return sols
        else:
            return sols[0]

    def detect_fragments(self, molecule: Chem.rdchem.Mol) -> dict:
        """Detect all the fragments in the molecule.

        Return a dictionary with the detected fragments as keys and a tuple
        with the atoms indexes of the fragment as values. For example, n-hexane
        for the UNIFAC model will return:

        .. code-block:: python

            {
                'CH3_0': (0,),
                'CH3_1': (5,),
                'CH2_0': (1,),
                'CH2_1': (2,),
                'CH2_2': (3,),
                'CH2_3': (4,)
            }

        You may note that multiple occurrence of a fragment name will be
        indexed. The convention is always: <fragment_name>_i where `i` is the
        index of the occurrence.

        Parameters
        ----------
        mol : Chem.rdchem.Mol
            Molecule to detect the fragments.

        Returns
        -------
        dict
            Detected fragments in the molecule.
        """
        detected_fragments = {
            f"{fragment_name}_{i}": atoms_tuple
            for fragment_name, mol in self.detection_mols.items()
            for i, atoms_tuple in enumerate(molecule.GetSubstructMatches(mol))
        }

        return detected_fragments



"""FragmentationResult class.

Base class to set fragmentation results. implements the drawing methods of the
solutions.
"""

import importlib

import numpy as np

from rdkit import Chem
from rdkit.Chem.Draw import rdMolDraw2D


class FragmentationResult:
    """Fragmentation result class.

    This class is used to store the fragmentation results and provide methods
    to visualize them.

    Parameters
    ----------
    molecule : Chem.rdchem.Mol
        Molecule to fragment.
    subgroups : dict
        Dictionary with the subgroups and the number of times they appear in
        the molecule.
    subgroups_atoms_indexes : dict
        Dictionary with the subgroups and the atoms indexes that belong to
        each subgroup.

    Attributes
    ----------
    molecule : Chem.rdchem.Mol
        Molecule to fragment.
    subgroups : dict
        Dictionary with the subgroups and the number of times they appear in
        the molecule.
    subgroups_atoms : dict
        Dictionary with the subgroups and the atoms indexes that belong to
        each subgroup.
    """

    def __init__(
        self,
        molecule: Chem.rdchem.Mol,
        subgroups: dict,
        subgroups_atoms_indexes: dict,
    ):
        self.molecule = molecule
        self.subgroups = subgroups
        self.subgroups_atoms = subgroups_atoms_indexes

    def get_solution_svg(
        self,
        title: str = "",
        width: float = 400,
        height: float = 200,
        title_font_size: float = 12,
        legend_font_size: float = 12,
        font: str = "Helvetica",
    ) -> str:
        """Create a SVG figure string of the fragmentation result.

        Parameters
        ----------
        title : str, optional
            Graph title, by default ""
        width : int, optional
            Graph width, by default 400
        height : int, optional
            Graph height, by default 200
        title_font_size : int, optional
            Font size of graph's title, by default 12
        legend_font_size : int, optional
            Legend font size, by default 12
        font : str, optional
            Text font, by default "Helvetica"

        Returns
        -------
        str
            SVG string.
        """
        # =====================================================================
        # Generate the colors for each subgroup (max 10 TODO)
        # =====================================================================
        how_many_subgroups = len(self.subgroups_atoms.keys())

        colors_rgb = self._generate_distinct_colors(how_many_subgroups)

        highlight = []
        atoms_colors = {}
        subgroups_colors = {}

        for idx, (subgroup, atoms) in enumerate(self.subgroups_atoms.items()):
            atms = np.array(atoms).flatten()

            subgroups_colors[subgroup] = colors_rgb[idx]
            highlight.extend(atms.tolist())

            for at in atms:
                atoms_colors[int(at)] = colors_rgb[idx]

        drawer = rdMolDraw2D.MolDraw2DSVG(width, height)

        drawer.DrawMolecule(
            self.molecule,
            highlightAtoms=highlight,
            highlightBonds=[],
            highlightAtomColors=atoms_colors,
        )
        drawer.FinishDrawing()
        svg = drawer.GetDrawingText().replace("svg:", "")

        # =====================================================================
        # Create legend
        # =====================================================================
        legend = ""

        for i, (name, color) in enumerate(subgroups_colors.items()):
            r, g, b, _ = color
            rect_color = f"rgb({int(r * 255)}, {int(g * 255)}, {int(b * 255)})"

            legend_name = name.replace("<", "&lt;").replace(">", "&gt;")

            # Color rectangle for the legend of the group
            legend += (
                f'<rect x="1" y="{5 + i * 25}" '
                f'width="{legend_font_size * 1.5}" '
                f'height="{legend_font_size * 1.5}" fill="{rect_color}" />'
            )

            # Group name and occurences for the legend
            legend += (
                f'<text x="{legend_font_size * 1.6}" y="{20 + i * 25}" '
                f'font-family="{font}" font-size="{legend_font_size}" '
                f'fill="black">{legend_name}: {self.subgroups[name]}</text>'
            )

        # =====================================================================
        # Create title
        # =====================================================================
        title = (
            f'<text x="{width/2}" y="40" '
            f'font-family="{font}" font-size="{title_font_size}" '
            f'font-weight="bold" fill="black" '
            f'text-anchor="middle">{title}</text>'
        )

        # =====================================================================
        # Set title and legend to figure
        # =====================================================================
        svg_complete = svg.replace("</svg>", f"{legend}{title}</svg>")

        return svg_complete

    def draw(
        self,
        title: str = "",
        width: float = 400,
        height: float = 200,
        title_font_size: float = 12,
        legend_font_size: float = 12,
        font: str = "Helvetica",
    ):
        """Create a IPython SVG object of the fragmentation result.

        This function is meant to be used in Jupyter notebooks to directly
        obtain the visual SVG figure on the notebook. It requires the IPython
        library to be installed.

        Parameters
        ----------
        title : str, optional
            Graph title, by default ""
        width : int, optional
            Graph width, by default 400
        height : int, optional
            Graph height, by default 200
        title_font_size : int, optional
            Font size of graph's title, by default 12
        legend_font_size : int, optional
            Legend font size, by default 12
        font : str, optional
            Text font, by default "Helvetica"

        Returns
        -------
        IPython.display.SVG
            SVG object of the fragmentation result.

        Raises
        ------
        ImportError
            IPython is not installed.
        """
        svg = self.get_solution_svg(
            title, width, height, title_font_size, legend_font_size, font
        )

        try:
            svg_lib = importlib.import_module("IPython.display")
        except ImportError:
            raise ImportError(
                "IPython is required to display the SVG. This function is"
                "meant to be used in Jupyter notebooks."
            )

        return svg_lib.SVG(svg)

    def _generate_distinct_colors(self, n: int) -> list:
        base_colors = np.array(
            [
                [0.12156863, 0.46666667, 0.70588235],  # blue
                [1.0, 0.49803922, 0.05490196],  # orange
                [0.17254902, 0.62745098, 0.17254902],  # green
                [0.83921569, 0.15294118, 0.15686275],  # red
                [0.58039216, 0.40392157, 0.74117647],  # purple
                [0.54901961, 0.3372549, 0.29411765],  # brown
                [0.89019608, 0.46666667, 0.76078431],  # pink
                [0.49803922, 0.49803922, 0.49803922],  # gray
                [0.7372549, 0.74117647, 0.13333333],  # yellow
                [0.09019608, 0.74509804, 0.81176471],  # cian
            ]
        )
        colors = [base_colors[i % len(base_colors)] for i in range(n)]
        return [(color[0], color[1], color[2], 0.65) for color in colors]



"""GibbsModel fragmentation module."""

from typing import List, Union

import pandas as pd

from rdkit import Chem

from ugropy.core.frag_classes.base.fragmentation_model import (
    FragmentationModel,
)
from ugropy.core.frag_classes.gibbs_model.gibbs_result import (
    GibbsFragmentationResult,
)
from ugropy.core.ilp_solvers.default_solver import DefaultSolver
from ugropy.core.ilp_solvers.ilp_solver import ILPSolver


class GibbsModel(FragmentationModel):
    """GibbsModel it's a fragmentation model dedicated to Gibbs excess models.

    unifac, psrk, dortmund are instances of this class.

    Parameters
    ----------
    subgroups : pd.DataFrame
        Model's subgroups. Index: 'group' (subgroups names). Mandatory columns:
        'smarts' (SMARTS representations of the group to detect its precense in
        the molecule).
    subgroups_info : Union[pd.DataFrame, None], optional
        Information of the model's subgroups (R, Q, subgroup_number,
        main_group), by default None

    Attributes
    ----------
    subgroups : pd.DataFrame
        Model's subgroups. Index: 'group' (subgroups names). Columns: 'smarts'
        (SMARTS representations of the group to detect its precense in the
        molecule).
    detection_mols : dict
        Dictionary cotaining all the rdkit Mol object from the detection_smarts
        subgroups column
    subgroups_info : pd.DataFrame
        Information of the model's subgroups. Columns: R, Q, subgroup_number,
        main_group. Index: 'group' (subgroups names)
    """

    def __init__(
        self,
        subgroups: pd.DataFrame,
        subgroups_info: Union[pd.DataFrame, None] = None,
        calculate_r_q: bool = True,
    ) -> None:
        super().__init__(
            subgroups=subgroups,
            allow_overlapping=False,
            fragmentation_result=GibbsFragmentationResult,
        )

        self._calculate_r_q = calculate_r_q

        # subgroups info
        if subgroups_info is None:
            self.subgroups_info = pd.DataFrame(
                [],
                columns=["group", "subgroup_number", "main_group", "R", "Q"],
            ).set_index("group")
        else:
            self.subgroups_info = subgroups_info

    def get_groups(
        self,
        identifier: Union[str, Chem.rdchem.Mol],
        identifier_type: str = "name",
        solver: ILPSolver = DefaultSolver,
        search_multiple_solutions: bool = False,
        search_nonoptimal: bool = False,
        solver_arguments: dict = {},
    ) -> Union[GibbsFragmentationResult, List[GibbsFragmentationResult]]:
        """Get the groups of a molecule.

        Parameters
        ----------
        identifier : Union[str, Chem.rdchem.Mol]
            Identifier of the molecule. You can use either the name of the
            molecule, the SMILEs of the molecule or a rdkit Mol object.
        identifier_type : str, optional
            Identifier type of the molecule. Use "name" if you are providing
            the molecules' name, "smiles" if you are providing the SMILES
            or "mol" if you are providing a rdkir mol object, by default "name"
        solver : ILPSolver, optional
            ILP solver class, by default DefaultSolver
        search_multiple_solutions : bool, optional
            Weather search for multiple solutions or not, by default False
            If False the return will be a FragmentationResult object, if True
            the return will be a list of FragmentationResult objects.
        search_nonoptimal : bool, optional
            If True, the solver will search for non-optimal solutions along
            with the optimal ones. This is useful when the user wants to find
            all possible combinations of fragments that cover the universe. By
            default False. If `search_multiple_solutions` is False, this
            parameter will be ignored.
        solver_arguments : dict, optional
            Dictionary with the arguments to be passed to the solver. For the
            DefaultSolver of ugropy you can change de PulP solver passing a
            dictionary like {"solver": "PULP_CBC_CMD"} and change the PulP
            solver. If empty it will use the default solver arguments, by
            default {}.

        Returns
        -------
        Union[GibbsFragmentationResult, List[GibbsFragmentationResult]]
            Fragmentation result. If search_multiple_solutions is False the
            return will be a FragmentationResult object, if True the return
            will be a list of FragmentationResult objects.
        """
        sol = super().get_groups(
            identifier,
            identifier_type,
            solver,
            search_multiple_solutions,
            search_nonoptimal,
            solver_arguments,
            subgroups_info=self.subgroups_info,
            calculate_r_q=self._calculate_r_q,
        )

        return sol



"""Fragmentation models implementations.

All models can be imported directly with:

.. code-block:: python

    from ugropy import joback, psrk, unifac, abdulelah_gani

You can check the group list and their SMARTS representation with:

.. code-block:: python

    joback.subgroups
    psrk.subgroups
    unifac.subgroups

Obtain the contribution of each group to the properties with:

.. code-block:: python

    joback.get_groups("n-hexane").subgroups
    unifac.get_groups("toluene").subgroups
    psrk.get_groups("CCO", "smiles").subgroups
    abdulelah_gani.get_groups("CCO", "smiles").primary.subgroups
    abdulelah_gani.get_groups("CCO", "smiles").secondary.subgroups
    abdulelah_gani.get_groups("CCO", "smiles").tertiary.subgroups

Attributes
----------
psrk: GibbsModel
    Predictive Soave-Redlich-Kwong FragmentationModel :cite:p:`ddbst, psrk1,
    psrk2`
unifac : GibbsModel
    Classic LV-UNIFAC FragmentationModel :cite:p:`ddbst, unifac1, unifac2,
    unifac3, unifac4, unifac5, unifac6`
dortmund : GibbsModel
    Dortmund mod-UNIFAC FragmentationModel :cite:p:`dor1, dor2, dor3, dor4,
    dor5, dor6, dor7, dor8, dor9, dor10, dor11`
joback: JobackModel
    Joback FragmentationModel :cite:p:`joback1, joback2`
abdulelah_gani_p: AbdulelahGaniPSTModel
    Abdulelah-Gani Primary Structures FragmentationModel :cite:p:`gani`
abdulelah_gani_s: AbdulelahGaniPSTModel
    Abdulelah-Gani Secondary Structures FragmentationModel :cite:p:`gani`
abdulelah_gani_t: AbdulelahGaniPSTModel
    Abdulelah-Gani Tertiary Structures FragmentationModel :cite:p:`gani`
abdulelah_gani: AbdulelahGaniModel
    Abdulelah-Gani properties estimation model :cite:p:`gani`
"""

from . import (
    abdulelah_gani_pmod,
    abdulelah_gani_smod,
    abdulelah_gani_tmod,
    dortmundmod,
    jobackmod,
    psrkmod,
    read_csv,
    unifacmod,
)


__all__ = [
    "abdulelah_gani_pmod",
    "abdulelah_gani_smod",
    "abdulelah_gani_tmod",
    "dortmundmod",
    "read_csv",
    "jobackmod",
    "psrkmod",
    "unifacmod",
]



"""Abdulelah-Gani fragmentation module :cite:p:`gani`."""

from typing import List, Union

import pandas as pd

from rdkit import Chem

from ugropy.core.frag_classes.abdulelah_gani.abdulelah_gani_pst import (
    AbdulelahGaniPSTModel,
)
from ugropy.core.frag_classes.abdulelah_gani.abdulelah_gani_result import (
    AGaniFragmentationResult,
)
from ugropy.core.ilp_solvers.default_solver import DefaultSolver
from ugropy.core.ilp_solvers.ilp_solver import ILPSolver


class AbdulelahGaniModel:
    """Abdulelah-Gani fragmentation model :cite:p:`gani`.

    Parameters
    ----------
    abdulelah_gani_p : AbdulelahGaniPSTModel
        The primary Abdulelah-Gani fragmentation model.
    abdulelah_gani_s : AbdulelahGaniPSTModel
        The secondary Abdulelah-Gani fragmentation model.
    abdulelah_gani_t : AbdulelahGaniPSTModel
        The tertiary Abdulelah-Gani fragmentation model.
    properties_contributions : pd.DataFrame
        The contributions parameters of each group for each property of the
        model.
    properties_biases : pd.DataFrame
        The biases parameters of each property of the model.

    Attributes
    ----------
    primary_model : AbdulelahGaniPSTModel
        The primary Abdulelah-Gani fragmentation model.
    secondary_model : AbdulelahGaniPSTModel
        The secondary Abdulelah-Gani fragmentation model.
    tertiary_model : AbdulelahGaniPSTModel
        The tertiary Abdulelah-Gani fragmentation model.
    properties_contributions : pd.DataFrame
        The contributions parameters of each group for each property of the
        model.
    properties_biases : pd.DataFrame
        The biases parameters of each property of the model
    """

    def __init__(
        self,
        abdulelah_gani_p: AbdulelahGaniPSTModel,
        abdulelah_gani_s: AbdulelahGaniPSTModel,
        abdulelah_gani_t: AbdulelahGaniPSTModel,
        properties_contributions: pd.DataFrame,
        properties_biases: pd.DataFrame,
    ) -> None:

        self.primary_model = abdulelah_gani_p
        self.secondary_model = abdulelah_gani_s
        self.tertiary_model = abdulelah_gani_t
        self.properties_contributions = properties_contributions
        self.properties_biases = properties_biases

    def get_groups(
        self,
        identifier: Union[str, Chem.rdchem.Mol],
        identifier_type: str = "name",
        solver: ILPSolver = DefaultSolver,
        search_multiple_solutions: bool = False,
        search_nonoptimal: bool = False,
        solver_arguments: dict = {},
    ) -> Union[AGaniFragmentationResult, List[AGaniFragmentationResult]]:
        """Get the groups of the molecule.

        Parameters
        ----------
        identifier : Union[str, Chem.rdchem.Mol]
            Identifier of the molecule. You can use either the name of the
            molecule, the SMILEs of the molecule or a rdkit Mol object.
        identifier_type : str, optional
            Identifier type of the molecule. Use "name" if you are providing
            the molecules' name, "smiles" if you are providing the SMILES
            or "mol" if you are providing a rdkir mol object, by default "name"
        solver : ILPSolver, optional
            ILP solver class, by default DefaultSolver
        search_multiple_solutions : bool, optional
            Weather search for multiple solutions or not, by default False
            If False the return will be a FragmentationResult object, if True
            the return will be a list of FragmentationResult objects.
        search_nonoptimal : bool, optional
            If True, the solver will search for non-optimal solutions along
            with the optimal ones. This is useful when the user wants to find
            all possible combinations of fragments that cover the universe. By
            default False. If `search_multiple_solutions` is False, this
            parameter will be ignored.
        solver_arguments : dict, optional
            Dictionary with the arguments to be passed to the solver. For the
            DefaultSolver of ugropy you can change de PulP solver passing a
            dictionary like {"solver": "PULP_CBC_CMD"} and change the PulP
            solver. If empty it will use the default solver arguments, by
            default {}.

        Returns
        -------
        Union[AGaniFragmentationResult, List[AGaniFragmentationResult]]
            Fragmentation result. If search_multiple_solutions is False the
            return will be a FragmentationResult object, if True the return
            will be a list of FragmentationResult objects.
        """
        primary_groups = self.primary_model.get_groups(
            identifier,
            identifier_type,
            solver,
            search_multiple_solutions,
            search_nonoptimal,
            solver_arguments,
        )

        secondary_groups = self.secondary_model.get_groups(
            identifier,
            identifier_type,
            solver,
            search_multiple_solutions=False,
        )

        tertiary_groups = self.tertiary_model.get_groups(
            identifier,
            identifier_type,
            solver,
            search_multiple_solutions=False,
        )

        if not search_multiple_solutions:
            result = AGaniFragmentationResult(
                primary_groups.molecule,
                primary_groups,
                secondary_groups,
                tertiary_groups,
                self.properties_contributions,
                self.properties_biases,
            )
        else:
            result = []

            for psol in primary_groups:
                result.append(
                    AGaniFragmentationResult(
                        psol.molecule,
                        psol,
                        secondary_groups,
                        tertiary_groups,
                        self.properties_contributions,
                        self.properties_biases,
                    )
                )

        return result



"""ugropy library.

A Python library designed to swiftly and effortlessly obtain the UNIFAC-like
groups from molecules by their names and subsequently integrate them into
inputs for thermodynamic libraries. UNIFAC, PSRK, Dortmund-UNIFAC,
Abdulelah-Gani, and Joback models are implemented.
"""

from .constants import ureg
from .core import instantiate_mol_object
from .core.frag_classes.abdulelah_gani.abdulelah_gani_pst import (
    AbdulelahGaniPSTModel,
)
from .core.frag_classes.abdulelah_gani.abdulelah_gani_pst_result import (
    AGaniPSTFragmentationResult,
)
from .core.frag_classes.base.fragmentation_model import (
    FragmentationModel,
)
from .core.frag_classes.base.fragmentation_result import (
    FragmentationResult,
)
from .core.frag_classes.gibbs_model.gibbs_model import GibbsModel
from .core.frag_classes.gibbs_model.gibbs_result import (
    GibbsFragmentationResult,
)
from .core.frag_classes.joback.joback_model import JobackModel
from .core.frag_classes.joback.joback_result import JobackFragmentationResult
from .core.ilp_solvers.default_solver import DefaultSolver
from .core.ilp_solvers.ilp_solver import ILPSolver
from .groups import Groups
from .models.abdulelah_gani_mod import abdulelah_gani
from .models.abdulelah_gani_pmod import abdulelah_gani_p
from .models.abdulelah_gani_smod import abdulelah_gani_s
from .models.abdulelah_gani_tmod import abdulelah_gani_t
from .models.dortmundmod import dortmund
from .models.jobackmod import joback
from .models.psrkmod import psrk
from .models.unifacmod import unifac


__all__ = [
    # =========================================================================
    # Misc
    # -------------------------------------------------------------------------
    "ureg",
    "writers",
    "instantiate_mol_object",
    "Groups",
    # =========================================================================
    # Fragmentation models classes
    # -------------------------------------------------------------------------
    "FragmentationModel",  # Base class
    "AbdulelahGaniPSTModel",
    "GibbsModel",
    "JobackModel",
    # =========================================================================
    # Fragmentation results classes
    # -------------------------------------------------------------------------
    "FragmentationResult",  # Base class
    "AGaniPSTFragmentationResult",
    "GibbsFragmentationResult",
    "JobackFragmentationResult",
    # =========================================================================
    # ILP Solvers
    # -------------------------------------------------------------------------
    "ILPSolver",  # Base class
    "DefaultSolver",
    # =========================================================================
    # Models instances
    # -------------------------------------------------------------------------
    "abdulelah_gani",
    "abdulelah_gani_p",
    "abdulelah_gani_s",
    "abdulelah_gani_t",
    "dortmund",
    "joback",
    "unifac",
    "psrk",
]



"""get_rdkit_object module."""

from functools import cache
from typing import Union

import pubchempy as pcp

from rdkit import Chem


def instantiate_mol_object(
    identifier: Union[str, Chem.rdchem.Mol], identifier_type: str = "name"
) -> Chem.rdchem.Mol:
    """Instantiate a RDKit Mol object from molecule's name, SMILES or mol.

    In case that the input its already a RDKit Mol object, the function will
    return the input object.

    Parameters
    ----------
    identifier : str or rdkit.Chem.rdchem.Mol
        Identifier of a molecule (name, SMILES or rdkit.Chem.rdchem.Mol).
        Example: hexane or CCCCCC for name or SMILES respectively.
    identifier_type : str, optional
        Use 'name' to search a molecule by name, 'smiles' to provide the
        molecule SMILES representation or 'mol' to provide a
        rdkit.Chem.rdchem.Mol object, by default "name".

    Returns
    -------
    rdkit.Chem.rdchem.Mol
        RDKit Mol object.
    """
    if identifier_type.lower() == "smiles":
        smiles = identifier
        chem_object = Chem.MolFromSmiles(smiles)

    elif identifier_type.lower() == "name":
        chem_object = instantiate_mol_from_name(identifier)

    elif identifier_type.lower() == "mol":
        chem_object = identifier

        if not isinstance(chem_object, Chem.rdchem.Mol):
            raise ValueError(
                "If 'mol' identifier type is used, the identifier must be a "
                "rdkit.Chem.Chem.rdchem.Mol object."
            )
    else:
        raise ValueError(
            f"Identifier type: {identifier_type} not valid, use: 'name', "
            "'smiles' or 'mol'"
        )

    return chem_object


@cache
def instantiate_mol_from_name(name: str) -> Chem.rdchem.Mol:
    """Instantiate a RDKit Mol object from molecule's name.

    The funcion uses `pubchempy` to get the molecule's SMILES representation
    from the molecule's name and then instantiate a RDKit Mol object.

    Parameters
    ----------
    name : str
        Name of the molecule.

    Returns
    -------
    rdkit.Chem.rdchem.Mol
        RDKit Mol object.
    """
    try:
        pcp_object = pcp.get_compounds(name, "name")[0]

        if pcp_object.canonical_smiles:
            chem_object = Chem.MolFromSmiles(pcp_object.canonical_smiles)
        elif pcp_object.inchi:
            chem_object = Chem.MolFromInchi(pcp_object.inchi)
        else:
            raise IndexError(
                f"Could not find a SMILES or InChI for the molecule '{name}' "
                "on PubChem."
            )
    except IndexError:
        raise ValueError(
            f"Could not find a molecule with the name '{name}' on " "PubChem"
        )

    return chem_object



"""ILPSolver abstract class.

Template for defining ILP solvers used by `ugropy` to solve the `Set Cover`
problem.
"""

import itertools
from abc import ABC, abstractmethod
from typing import List


class ILPSolver(ABC):
    """ILPSolver abstract class.

    Template for defining ILP solvers used by `ugropy` to solve the `Set Cover`
    problem.

    To inherit from this class, the following methods must be implemented:

    - `solve_one_problem`:
        Method that solves one instance of the `Set Cover` problem. This method
        must return a list of integers representing the chosen fragments. Also
        must append the solution to the `selected_fragments` attribute as a
        dictionary.
    - `solve`:
        This class call the `solve_one_problem` method multiples times if the
        `search_multiple_solutions` attribute is True. The successives calls
        must include the previous solutions obtained as restrictions.

    You can personalize both methods as you want. The library will call the
    `solve` method and at the end of its execution the `selected_fragments`
    attribute must contain the selected fragments for each solution.

    Parameters
    ----------
    overlapped_atoms : List[int]
        List of atoms that are overlapped. On the `Set Cover` problem therms,
        these atoms represents the universe of the problem.
    fragments : dict[str: List[int]]
        Dictionary with the fragments (groups) to be selected. On the `Set
        Cover` problem therms, these fragments represents the sets to be
        selected to cover all the universe. The key is the name of the fragment
        and the value is a list of atoms that compose the fragment. Some atoms
        that represents a fragment could not be present in the universe. This
        is handled by adding those "free atoms" to the universe, forcing that
        fragment to appear on the solution. This updating of the universe is
        done on the `__init__` method, so you don't need to worry about it.
    search_multiple_solutions : bool, optional
        If True, the solver will search for multiple solutions, by default
        False
    search_nonoptimal : bool, optional
        If True, the solver will search for non-optimal solutions along with
        the optimal ones. This is useful when the user wants to find all
        possible combinations of fragments that cover the universe. By default
        False. If `search_multiple_solutions` is True, this parameter will be
        ignored.
    solver_arguments : dict, optional
        Dictionary with the arguments to be passed to the solver.

    Attributes
    ----------
    overlapped_atoms : List[int]
        List of atoms that are overlapped. On the `Set Cover` problem therms,
        these atoms represents the universe of the problem.
    fragments : dict[str: List[int]]
        Fragments to choose to cover the universe.
    search_multiple_solutions : bool
        If True, the solver will search for multiple solutions.
    search_nonoptimal : bool
        If True, the solver will search for non-optimal solutions along with
        the optimal ones.
    selected_fragments : List[List[str]]
        Solutions found by the `solve` method. Its a list  that contains the
        different obtained solutions as lists of strings. Each string
        corresponds to the name of the selected fragments.
    overlapped_fragments : dict
        Fragments that participates on overlapped atoms.
    universe : set
        Set with all elements that must be covered by the fragments. This set
        is composed by the overlapped atoms and the atoms that are present in
        the overlapped fragments but not in the overlapped atoms.
    """

    def __init__(
        self,
        overlapped_atoms: List,
        fragments: dict,
        search_multiple_solutions: bool = False,
        search_nonoptimal: bool = False,
        solver_arguments: dict = {},
    ):
        self.overlapped_atoms = overlapped_atoms
        self.fragments = fragments
        self.search_multiple_solutions = search_multiple_solutions
        self.search_nonoptimal = search_nonoptimal

        if solver_arguments == {}:
            self.solver_arguments = {
                "solver": "PULP_CBC_CMD",
            }
        else:
            self.solver_arguments = solver_arguments

        # Attribute that will store the selected fragments after a
        # `solve` method call.
        self.selected_fragments = []

        # Get overlapped fragments (which participates on overlapped atoms)
        self.overlapped_fragments = {}

        for name, fragment in self.fragments.items():
            if any(atom in self.overlapped_atoms for atom in fragment):
                self.overlapped_fragments[name] = fragment

        # set universe
        self.universe = set(self.overlapped_atoms)

        # Update universe with all elements in the overlapped fragments that
        # not present in the overlapped atoms but participates in a fragment.
        all_elements = set(
            itertools.chain.from_iterable(self.overlapped_fragments.values())
        )

        self.universe.update(all_elements)

    @abstractmethod
    def solve_one_problem(self, not_valid_solutions: List) -> List[int]:
        """Solves one instance of the `Set Cover` problem.

        Parameters
        ----------
        not_valid_solutions : List[list[int]]
            List with binary values representing the fragments that are choosen
            with a 1 and the not choosen with a 0. This list contains the lists
            of the solutions already found.

        Returns
        -------
        List[int]
            Solution found by the solver. The list must contain binary values
            representing the fragments that are choosen with a 1 and the not
            choosen with a 0. Of course, the length of the list must be equal
            to the lenght of the `overlapped_fragments` attribute.

        Raises
        ------
        NotImplementedError
            Abstract method not implemented
        """
        raise NotImplementedError(
            "Abstract method `solve_one_problem` not implemented"
        )

    @abstractmethod
    def solve(self) -> None:
        """Solve method called by fragmentations models.

        Ath the end of this method's execution, the `selected_fragments` must
        contain the selected fragments (solutions) as a list of dictionaries.

        Raises
        ------
        NotImplementedError
            Abstract method not implemented
        """
        raise NotImplementedError("Abstract method `solve` not implemented")



"""JobackFragmentationResult module."""

from typing import Union

import numpy as np

import pandas as pd

import pint

from rdkit import Chem
from rdkit.Chem import Descriptors

from ugropy.constants import R, ureg
from ugropy.core.frag_classes.base.fragmentation_result import (
    FragmentationResult,
)


class JobackFragmentationResult(FragmentationResult):
    """Joback group contribution properties estimator.

    The class recieves either the Joback and Reid model's :cite:p:`joback1,
    joback2` groups, name or smiles of a molecule and estimates the its
    properties.

    Parameters
    ----------
    molecule : Chem.rdchem.Mol
        RDKit molecule object.
    subgroups : dict
        Dictionary of subgroups.
    subgroups_atoms_indexes : dict
        Dictionary of subgroups atoms indexes.
    properties_contributions : pd.DataFrame
        DataFrame with Joback's properties contributions.
    normal_boiling_point : float, optional
        User provided experimental normal boiling point [K].

    Attributes
    ----------
    subgroups : dict
        Joback functional groups of the molecule.
    experimental_boiling_temperature : pint.Quantity
        User provided experimental normal boiling point [K].
    critical_temperature : pint.Quantity
        Joback estimated critical temperature [K].
    critical_pressure : pint.Quantity
        Joback estimated critical pressure [bar].
    critical_volume : pint.Quantity
        Joback estimated critical volume [cm³/mol].
    normal_boiling_point : pint.Quantity
        Joback estimated normal boiling point [K].
    fusion_temperature : pint.Quantity
        Joback estimated fusion temperature [K].
    ig_formation_formation : pint.Quantity
        Joback estimated enthalpy of formation ideal gas at 298 K [kJ/mol].
    ig_gibbs_formation : pint.Quantity
        Joback estimated Gibbs energy of formation ideal gas at 298 K [K].
    heat_capacity_ideal_gas_params : dict
        Joback estimated Reid's ideal gas heat capacity equation parameters
        [J/mol/K].
    fusion_enthalpy : pint.Quantity
        Joback estimated fusion enthalpy [kJ/mol].
    vaporization_enthalpy : pint.Quantity
        Joback estimated vaporization enthalpy at the normal boiling point
        [kJ/mol].
    sum_na : float
        Joback n_A contribution to liquid viscosity [Pa s].
    sum_nb : float
        Joback n_B contribution to liquid viscosity [Pa s].
    molecular_weight : pint.Quantity
        Molecular weight from Joback's subgroups [g/mol].
    acentric_factor : pint.Quantity
        Acentric factor from Lee and Kesler's equation :cite:p:`joback1`.
    vapor_pressure_params : dict
        Vapor pressure G and k parameters for the Riedel-Plank-Miller
        equation [bar] :cite:p:`joback1`.
    """

    def __init__(
        self,
        molecule: Chem.rdchem.Mol,
        subgroups: dict,
        subgroups_atoms_indexes: dict,
        properties_contributions: pd.DataFrame,
        normal_boiling_point: float = None,
    ) -> None:
        # Initialize the FragmentationResult attributes
        super().__init__(molecule, subgroups, subgroups_atoms_indexes)

        # experimental boiling temperature
        self.experimental_boiling_temperature = normal_boiling_point

        # Original Joback properties
        self.critical_temperature: pint.Quantity = None
        self.critical_pressure: pint.Quantity = None
        self.critical_volume: pint.Quantity = None
        self.normal_boiling_point: pint.Quantity = None
        self.fusion_temperature: pint.Quantity = None
        self.ig_enthalpy_formation: pint.Quantity = None
        self.ig_gibbs_formation: pint.Quantity = None
        self.heat_capacity_ideal_gas_params = np.array([])
        self.fusion_enthalpy: pint.Quantity = None
        self.vaporization_enthalpy: pint.Quantity = None
        self.sum_na = None
        self.sum_nb = None
        self.molecular_weight: pint.Quantity = None

        # Extra properties
        self.acentric_factor: pint.Quantity = None
        self.vapor_pressure_params = {}

        # Fill the properties' values
        if self.subgroups != {}:
            self._calculate_properties(properties_contributions)

    def heat_capacity_ideal_gas(
        self, temperature: Union[float, np.ndarray]
    ) -> pint.Quantity:
        """Calculate the ideal gas heat capacity [J/mol/K].

        Uses the Joback estimated Reid's ideal gas heat capacity equation
        parameters [J/mol/K].

        Parameters
        ----------
        temperature : Union[float, np.ndarray]
            Temperature [K]

        Returns
        -------
        Union[float, np.ndarray]
            Ideal gas heat capacity [J/mol/K].
        """
        a, b, c, d = self.heat_capacity_ideal_gas_params

        t = temperature

        return (a + b * t + c * t**2 + d * t**3) * ureg.J / ureg.mol / ureg.K

    def heat_capacity_liquid(
        self, temperature: Union[float, np.ndarray]
    ) -> pint.Quantity:
        """Calculate the liquid heat capacity [J/mol/K].

        Uses the Rowlinson-Bondi :cite:p:`joback1` equation with the Joback
        estimated properties.

        Parameters
        ----------
        temperature : Union[float, np.ndarray]
            Temperature [K]

        Returns
        -------
        Union[float, np.ndarray]
            Ideal gas heat capacity [J/mol/K].
        """
        tr = temperature / self.critical_temperature.magnitude
        w = self.acentric_factor.magnitude

        c_p0 = self.heat_capacity_ideal_gas(temperature).magnitude

        c_pl = c_p0 + R * (
            2.56
            + 0.436 * (1 - tr) ** (-1)
            + w
            * (
                2.91
                + 4.28 * (1 - tr) ** (-1 / 3) * tr ** (-1)
                + 0.296 * (1 - tr) ** (-1)
            )
        )

        return c_pl * ureg.J / ureg.mol / ureg.K

    def viscosity_liquid(
        self, temperature: Union[float, np.ndarray]
    ) -> pint.Quantity:
        """Calculate the Joback estimated liquid viscosity [Pa s].

        Parameters
        ----------
        temperature : Union[float, np.ndarray]
            Temperature [K]

        Returns
        -------
        Union[float, np.ndarray]
            Liquid viscosity [Pa s].
        """
        t = temperature

        n_l = self.molecular_weight.magnitude * np.exp(
            (self.sum_na - 597.82) / t + self.sum_nb - 11.202
        )
        return n_l * ureg.Pa * ureg.s

    def vapor_pressure(
        self, temperature: Union[float, np.ndarray]
    ) -> pint.Quantity:
        """Calculate the vapor pressure [bar].

        Uses the Riedel-Plank-Miller :cite:p:`joback1` equation with the Joback
        estimated properties.

        Parameters
        ----------
        temperature : Union[float, np.ndarray]
            Temperature [K]

        Returns
        -------
        Union[float, np.ndarray]
            Vapor pressure [bar]
        """
        tr = temperature / self.critical_temperature.magnitude

        g = self.vapor_pressure_params["G"]
        k = self.vapor_pressure_params["k"]

        vp_r = np.exp(-g / tr * (1 - tr**2 + k * (3 + tr) * (1 - tr) ** 3))

        vp = vp_r * self.critical_pressure.magnitude

        return vp * ureg.bar

    def _calculate_properties(self, contribs: pd.DataFrame) -> None:
        """Obtain the molecule properties from Joback's groups."""
        groups = list(self.subgroups.keys())
        ocurr = list(self.subgroups.values())

        df = contribs.loc[groups]

        # =====================================================================
        # Calculate complete contribution properties (no contribution missing)
        # =====================================================================
        tc_c = df["Tc"].to_numpy()
        pc_c = df["Pc"].to_numpy()
        vc_c = df["Vc"].to_numpy()
        tb_c = df["Tb"].to_numpy()
        tf_c = df["Tf"].to_numpy()
        hform_c = df["Hform"].to_numpy()
        gform_c = df["Gform"].to_numpy()
        hvap_c = df["Hvap"].to_numpy()
        numa_c = df["num_of_atoms"].to_numpy()
        mw_c = Descriptors.MolWt(self.molecule)

        # Molecular weight
        self.molecular_weight = mw_c * ureg.g / ureg.mol

        # Joback normal boiling point (Tb)
        self.normal_boiling_point = (198.2 + np.dot(ocurr, tb_c)) * ureg.K

        # Fusion temperature (Tf)
        self.fusion_temperature = (122.5 + np.dot(ocurr, tf_c)) * ureg.K

        # Used normal boiling point for calculations
        if self.experimental_boiling_temperature is not None:
            tb = self.experimental_boiling_temperature
        else:
            tb = self.normal_boiling_point.magnitude

        # Critical temperature (Tc) normal boiling temperature for calculations
        self.critical_temperature = (
            tb
            * (
                0.584
                + 0.965 * np.dot(ocurr, tc_c)
                - (np.dot(ocurr, tc_c)) ** 2
            )
            ** (-1)
            * ureg.K
        )

        # Critical pressure (Pc)
        self.critical_pressure = (
            0.113 + 0.0032 * np.dot(ocurr, numa_c) - np.dot(ocurr, pc_c)
        ) ** (-2) * ureg.bar

        # Critical volume (Vc)
        self.critical_volume = (
            (17.5 + np.dot(ocurr, vc_c)) * ureg.cm**3 / ureg.mol
        )

        # Standard enthalpy of formation (298 K)
        self.ig_enthalpy_formation = (
            (68.29 + np.dot(ocurr, hform_c)) * ureg.kJ / ureg.mol
        )

        # Standard Gibbs energy of formation (298 K)
        self.ig_gibbs_formation = (
            (53.88 + np.dot(ocurr, gform_c)) * ureg.kJ / ureg.mol
        )

        # Enthalpy of vaporization
        self.vaporization_enthalpy = (
            (15.30 + np.dot(ocurr, hvap_c)) * ureg.kJ / ureg.mol
        )

        # =====================================================================
        # Incomplete contribution properties (some contribution missing)
        # =====================================================================
        # Heat capacity
        if "-N= (ring)" not in groups:
            a_c = df["a"].to_numpy()
            b_c = df["b"].to_numpy()
            c_c = df["c"].to_numpy()
            d_c = df["d"].to_numpy()

            a = np.dot(ocurr, a_c) - 37.93
            b = np.dot(ocurr, b_c) + 0.21
            c = np.dot(ocurr, c_c) - 3.91e-4
            d = np.dot(ocurr, d_c) + 2.06e-7

            self.heat_capacity_ideal_gas_params = np.array([a, b, c, d])

        # Enthalpy of fusion
        if all(df["Hfusion"].notnull()):
            hfusion_c = df["Hfusion"].to_numpy()
            self.fusion_enthalpy = (
                (-0.88 + np.dot(ocurr, hfusion_c)) * ureg.kJ / ureg.mol
            )

        # Liquid viscosity
        if all(df["na"].notnull()):
            na_c = df["na"].to_numpy()
            nb_c = df["nb"].to_numpy()

            self.sum_na = np.dot(ocurr, na_c)
            self.sum_nb = np.dot(ocurr, nb_c)

        # =====================================================================
        # Extra properties
        # =====================================================================
        # Reduced normal boiling point temperature
        t_br = tb / self.critical_temperature.magnitude

        # Lee and Kesler's equation (acentric factor)
        pc = self.critical_pressure.magnitude

        self.acentric_factor = (
            (
                -np.log(pc)
                - 5.92714
                + 6.09648 / t_br
                + 1.28862 * np.log(t_br)
                - 0.169347 * t_br**6
            )
            / (
                15.2518
                - 15.6875 / t_br
                - 13.4721 * np.log(t_br)
                + 0.43577 * t_br**6
            )
            * ureg.dimensionless
        )

        # Riedel-Plank-Miller equation (vapor pressure [bar])
        h = t_br * np.log(pc / 1.01325) / (1 - t_br)

        g = 0.4835 + 0.4605 * h

        k = (h / g - (1 + t_br)) / ((3 + t_br) * (1 - t_br) ** 2)

        self.vapor_pressure_params = {"G": g, "k": k}



"""to_thermo module."""

from ugropy.core.frag_classes.gibbs_model.gibbs_model import GibbsModel


def to_thermo(mol_subgroups: dict, model: GibbsModel) -> dict:
    """Obtain the subgroups dictionary to the Caleb Bell's Thermo library.

    Thermo: https://github.com/CalebBell/thermo

    Parameters
    ----------
    mol_subgroups : dict
        ugropy subgroups.
    model : GibbsModel
        Gibbs excess FragmentationModel (unifac, psrk, etc).

    Returns
    -------
    dict
        Thermo fragmentation subgroups.
    """
    thermo_groups = {}
    for group, occurrence in mol_subgroups.items():
        group_num = int(model.subgroups_info.loc[group, "subgroup_number"])

        thermo_groups[group_num] = occurrence

    return thermo_groups



"""constants module.

Attributes
----------
R : float
    Gas constant [J/mol/K]
ureg : pint.UnitRegistry
    Unit registry of Pint library
"""

from pathlib import Path

from pint import UnitRegistry


# Unit registry
ureg = UnitRegistry()


# constants.py path
_here = Path(__file__).parent

# CSVs path
_csvs = _here / "groupscsv"

# Gas constant [J/mol/K]
R = 8.31446261815324



"""Gibbs fragmentation result module."""

import pandas as pd

from rdkit import Chem

from ugropy.core.frag_classes.base.fragmentation_result import (
    FragmentationResult,
)


class GibbsFragmentationResult(FragmentationResult):
    """Gibbs fragmentation result class.

    Parameters
    ----------
    molecule : Chem.rdchem.Mol
        RDKit molecule object.
    subgroups : dict
        Dictionary of subgroups.
    subgroups_atoms_indexes : dict
        Dictionary of subgroups atoms indexes.
    subgroups_info : pd.DataFrame
        DataFrame with subgroups information.
    calculate_r_q : bool
        If True, calculate R and Q values for the molecule.
    calculate_num_dict : bool, optional
        If True, calculate the subgroup numbers dictionary. Default is True.

    Attributes
    ----------
    molecule : Chem.rdchem.Mol
        Molecule to fragment.
    subgroups : dict
        Dictionary with the subgroups and the number of times they appear in
        the molecule.
    subgroups_atoms : dict
        Dictionary with the subgroups and the atoms indexes that belong to
        each subgroup.
    r : float
        Gibss excess model R value estimation of the molecule.
    q : float
        Gibss excess model Q value estimation of the molecule.
    subgroups_num : dict
        Dictionary with the subgroup numbers and the number of times they
        appear in the molecule.
    """

    def __init__(
        self,
        molecule: Chem.rdchem.Mol,
        subgroups: dict,
        subgroups_atoms_indexes: dict,
        subgroups_info: pd.DataFrame,
        calculate_r_q: bool,
        calculate_num_dict: bool = True,
    ):
        super().__init__(molecule, subgroups, subgroups_atoms_indexes)

        r = 0.0
        q = 0.0

        # R and Q
        if self.subgroups != {}:
            if calculate_r_q:
                for group, n in self.subgroups.items():
                    r += n * subgroups_info.loc[group, "R"]
                    q += n * subgroups_info.loc[group, "Q"]

                self.r = r
                self.q = q
            else:
                self.r = None
                self.q = None
        else:
            self.r = None
            self.q = None

        # Subgroups numbers dictionary
        if self.subgroups != {}:
            if calculate_num_dict:
                self.subgroups_num = {}

                for group, occ in self.subgroups.items():
                    snum = int(subgroups_info.loc[group, "subgroup_number"])

                    self.subgroups_num[snum] = occ

            else:
                self.subgroups_num = None



"""Default ILP solver for the set cover problem.

The DefaultSolver uses the Pulp library to solve the set cover problem.
"""

from typing import List

import numpy as np

import pulp

from ugropy.core.ilp_solvers.ilp_solver import ILPSolver


class DefaultSolver(ILPSolver):
    """Default ILP solver for the set cover problem."""

    def solve_one_problem(self, not_valid_solutions: List = []) -> List[int]:
        """Solve one `Set Cover` problem with the Pulp library.

        Parameters
        ----------
        not_valid_solutions : List, optional
            Already got solutions, by default []

        Returns
        -------
        List[int]
            List of binary values representing the selected fragments.
        """
        # =====================================================================
        # ILP formulation (Pulp)
        # =====================================================================
        problem = pulp.LpProblem("Set_Cover_Problem", pulp.LpMinimize)

        n_frag = len(self.overlapped_fragments)

        # Decision variables
        x = pulp.LpVariable.dicts("x", range(n_frag), cat="Binary")

        # Objective function
        problem += pulp.lpSum([x[i] for i in range(n_frag)])

        # Constraints 1: each element must be covered by only one subset
        for elem in self.universe:
            sum_list = []
            for i, subset in enumerate(self.overlapped_fragments.values()):
                if elem in subset:
                    sum_list.append(x[i])

            problem += pulp.lpSum(sum_list) == 1

        # Constaints 2: if are not_valid_solutions, the new ones cannot use
        # less elements than the previous ones, otherwise the first one was not
        # optimal.
        if not_valid_solutions:
            opt = np.sum(np.array(not_valid_solutions[0], dtype=int))

            if not self.search_nonoptimal:
                # Do you want to search only optimals solutions?
                problem += pulp.lpSum([x[i] for i in range(n_frag)]) == opt

        # Constaints 3: already got solutions not allowed
        for solution in not_valid_solutions:
            problem += (
                pulp.lpSum([x[i] * solution[i] for i in range(n_frag)])
                <= pulp.lpSum([solution[i] for i in range(n_frag)]) - 1
            )

        # Solver configuration verbosity
        solver = pulp.getSolver(self.solver_arguments["solver"], msg=False)

        # Solve
        problem.solve(solver)

        # Selected fragments (solution)
        if pulp.LpStatus[problem.status] == "Optimal":
            selected_subsets = [
                name
                for i, name in enumerate(self.overlapped_fragments.keys())
                if pulp.value(x[i]) == 1
            ]

            self.selected_fragments.append(selected_subsets)

            return [pulp.value(x[i]) for i in range(n_frag)]
        else:
            # No feasible problem
            return None

    def solve(self) -> None:
        """Run multiple times the solve_one_problem method."""
        # Multiple solutions
        if self.search_multiple_solutions:
            not_valid_solutions = []
            while True:
                solution = self.solve_one_problem(not_valid_solutions)
                if solution is None:
                    break
                not_valid_solutions.append(solution)
        else:
            _ = self.solve_one_problem()



"""Joback fragmentation module."""

from typing import List, Union

import pandas as pd

from rdkit import Chem

from ugropy.core.frag_classes.base.fragmentation_model import (
    FragmentationModel,
)
from ugropy.core.frag_classes.joback.joback_result import (
    JobackFragmentationResult,
)
from ugropy.core.ilp_solvers.default_solver import DefaultSolver
from ugropy.core.ilp_solvers.ilp_solver import ILPSolver


class JobackModel(FragmentationModel):
    """Joback Fragmentation model dedicated to properties estimation models.

    Parameters
    ----------
    subgroups : pd.DataFrame
        Model's subgroups. Index: 'group' (subgroups names). Mandatory columns:
        'smarts' (SMARTS representations of the group to detect its precense in
        the molecule).
    properties_contributions : pd.DataFrame, optional
        Group's properties contributions, by default None.

    Attributes
    ----------
    subgroups : pd.DataFrame
        Model's subgroups. Index: 'group' (subgroups names). Columns: 'smarts'
        (SMARTS representations of the group to detect its precense in the
        molecule).
    detection_mols : dict
        Dictionary cotaining all the rdkit Mol object from the detection_smarts
        subgroups column
    properties_contributions : pd.DataFrame
        Group's properties contributions.
    """

    def __init__(
        self,
        subgroups: pd.DataFrame,
        properties_contributions: Union[pd.DataFrame, None],
    ) -> None:

        super().__init__(
            subgroups=subgroups,
            allow_overlapping=False,
            fragmentation_result=JobackFragmentationResult,
        )

        # Properties
        self.properties_contributions = properties_contributions

    def get_groups(
        self,
        identifier: Union[str, Chem.rdchem.Mol],
        identifier_type: str = "name",
        solver: ILPSolver = DefaultSolver,
        search_multiple_solutions: bool = False,
        search_nonoptimal: bool = False,
        solver_arguments: dict = {},
        normal_boiling_point: float = None,
    ) -> Union[JobackFragmentationResult, List[JobackFragmentationResult]]:
        """Get Jobacks groups from a molecule.

        Parameters
        ----------
        identifier : Union[str, Chem.rdchem.Mol]
            Identifier of the molecule. You can use either the name of the
            molecule, the SMILEs of the molecule or a rdkit Mol object.
        identifier_type : str, optional
            Identifier type of the molecule. Use "name" if you are providing
            the molecules' name, "smiles" if you are providing the SMILES
            or "mol" if you are providing a rdkir mol object, by default "name"
        solver : ILPSolver, optional
            ILP solver class, by default DefaultSolver
        search_multiple_solutions : bool, optional
            Weather search for multiple solutions or not, by default False
            If False the return will be a FragmentationResult object, if True
            the return will be a list of FragmentationResult objects.
        search_nonoptimal : bool, optional
            If True, the solver will search for non-optimal solutions along
            with the optimal ones. This is useful when the user wants to find
            all possible combinations of fragments that cover the universe. By
            default False. If `search_multiple_solutions` is False, this
            parameter will be ignored.
        normal_boiling_point : float, optional
            Experimental normal boiling point of the molecule on Kelvin. Its
            used to improve the properties calculations. Joback uses the
            estimated normal boiling point if no provided, by default None
        solver_arguments : dict, optional
            Dictionary with the arguments to be passed to the solver. For the
            DefaultSolver of ugropy you can change de PulP solver passing a
            dictionary like {"solver": "PULP_CBC_CMD"} and change the PulP
            solver. If empty it will use the default solver arguments, by
            default {}.

        Returns
        -------
        Union[JobackFragmentationResult, List[JobackFragmentationResult]]
            Fragmentation result. If search_multiple_solutions is False the
            return will be a FragmentationResult object, if True the return
            will be a list of FragmentationResult objects.
        """
        sol = super().get_groups(
            identifier=identifier,
            identifier_type=identifier_type,
            solver=solver,
            search_multiple_solutions=search_multiple_solutions,
            search_nonoptimal=search_nonoptimal,
            solver_arguments=solver_arguments,
            normal_boiling_point=normal_boiling_point,
            properties_contributions=self.properties_contributions,
        )

        return sol



"""Classic Liquid Vapor UNIFAC FragmentationModel implementation.

Import and use the UNIFAC FragmentationModel with:

.. code-block:: python

    from ugropy import unifac

    # Get groups from molecule's name
    tol = unifac.get_groups("toluene")

    print(tol.subgroups)

    # Get groups from molecule's SMILES
    eth = unifac.get_groups("CCO", "smiles")

    print(eth.subgroups)

Attributes
----------
unifac : GibbsModel
    Classic LV-UNIFAC FragmentationModel :cite:p:`ddbst, unifac1, unifac2,
    unifac3, unifac4, unifac5, unifac6`
"""

from ugropy.constants import _csvs
from ugropy.core.frag_classes.gibbs_model.gibbs_model import GibbsModel
from ugropy.models.read_csv import _rd


# =============================================================================
# LV-UNIFAC
# =============================================================================
_uni = _csvs / "unifac"

_uni_sg = _rd(_uni / "unifac_subgroups.csv", "group")
_uni_info = _rd(_uni / "unifac_info.csv", "group")

unifac = GibbsModel(subgroups=_uni_sg, subgroups_info=_uni_info)
